######
#
# 
# Python code used to solve differential equations
#   
#
#####

################################# Start-up ####################################
## Modules
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.axes as axes
import math
from scipy.integrate import ode
import csv
import pylab

## values used later
limit = 0.1
timef = 63.0
fontSize = 16

## Initial conditions
S = 950 # external solution K conc. (uM)
Wp0 = 0.15 # g
cr0 = 56 # umol/g
cs0 = 102 # umol/g
t0 = 0.0 # days
dt = 1.0 # days

## plotting
pylab.rcParams['figure.figsize'] = (13, 7)

################### Parameters ################################################
# "Genome"
## parameter values
wtG = [0.17, # root:plant volume ratio, delta
       58.9,63, # uptake: mu, kappa
       11.8,29,2.8, # net translocation: chi, phi, theta
       0.129,16,1, # root growth: alpha_root, beta_root
       0.105,40,3.2, # shoot growth: alpha_shoot, beta_shoot, sigma
       4, 37, 12, # yield: eta, nu, lambda
       Wp0, cr0, cs0] # initial: omega, gammaR, gammaS
# 
## parameter names
params = {0:r"$\delta$",
          1:r"$\mu$", 2:r"$\kappa$",
          3:r"$\chi$", 4:r"$\phi$", 5:r"$\theta$",
          6:r"$\alpha_{root}$", 7:r"$\beta_{root}$", 8:r"$\sigma_{root}$",
          9:r"$\alpha_{shoot}$", 10:r"$\beta_{shoot}$",11:r"$\sigma_{shoot}$",
          12:r"$\eta$", 13:r"$\nu$", 14:r"$\lambda$",
          15:r"$\omega$", 16:r"$\gamma_{root}$", 17:r"$\gamma_{shoot}$"}

######################### Functions useful for analyses #######################
def mutate(initialGenome, position=None, by=None, originalGenome=None, limit=None):
    """Mutate a genome at given position by a certain amount"""
    if by == None: ## if mutation doesn't change the genome at all
        return(initialGenome) ## return initial genome unchange
        
    else: ## otherwise the genome is changed
        newGenome = initialGenome.copy() # make a copy of the initial genome
        newGenome[position] = newGenome[position] * by # change the gene at a given position in the genome by a given amount
        
        if limit != None: # if there is a limit to the size of change a gene can undergo
        
            if newGenome[position] > (1+limit) * originalGenome[position]: # if the mutation makes the gene value bigger than the limit
                newGenome[position] = originalGenome[position] * (1+limit) # set the mutated value to the limit
                
            elif newGenome[position] < (1-limit) * originalGenome[position]: # or if the mutated gene value is less than the limit
                newGenome[position] = originalGenome[position] * (1-limit) # set it to the lower limit

            return (newGenome) # then return the new genome that does not violate the limits
        
        else: # if the limits were not exceeded by the mutation
            return(newGenome) # return the mutated genome

def decimalRange(start, stop, step): # python range only uses integers. Decimals can be useful
    """Returns a decimal range between start and stop with gap of size step"""
    i = start
    while i <= stop:
        yield round(i,2)
        i = i + step
        
def wCSV(geneNo=None, elast=None, name="output.csv"):
    with open(name, "w") as f:
        d = [geneNo,elast]
        writer = csv.writer(f)
        writer.writerows(d)
# run with: e.g. wCSV(geneNo=gno, elast=e, name="rgrRatio.csv")

#################################### ODE solver ###############################
def solver(g=wtG, S=S, report=True, tfinal=timef):
    Wr0 = g[0]*g[15] # iniital root mass
    Ws0 = (1-g[0])*g[15]  # initial shoot mass
    Qr0 =  g[16] * Wr0 # initial root content
    Qs0 = g[17] * Ws0 # initial shoot content
    Y0 = np.array([Wr0,Ws0,Qr0,Qs0]) # initial plant conditions
    def f(t,Y):
        ## unpack the genome
        delta, mu,kappa, chi,phi,theta, alphaR, betaR, sigmaR, \
            alphaS, betaS, sigmaS, eta, nu, Lambda, omega, gammaR, gammaS = g[0],g[1],g[2],g[3],g[4],g[5],g[6],g[7],\
            g[8],g[9],g[10],g[11],g[12], g[13], g[14], g[15], g[16], g[17]
        ## unpack the state of the plant
        Wr,Ws,Qr,Qs = Y
        cr = Qr/Wr # calcualte cr
        cs = Qs/Ws # calculate 
        ## sanity checking
        if cr < 0:
            print ("NEGATIVE ROOT CONCENTRATION !!!!!")
            dWrdt,dWsdt,dQrdt,dQsdt = [0,0,0,0]
        elif cs < 0:
            print ("NEGATIVE SHOOT CONCENTRATION !!!!!")
            dWrdt,dWsdt,dQrdt,dQsdt = [0,0,0,0]
        else: # if things make sense (no negative concs.) solve the system
            u = ((mu*S) / (kappa+S)) * Wr # absorption 
            T = ((chi*(cr**theta)) / ((phi**theta)+(cr**theta))) * Ws # translocation
            dQrdt = u - T
            dQsdt = T
        
            dWrdt = ( (alphaR * (cr**sigmaR) )/( (betaR**sigmaR) + (cr**sigmaR) ) ) * Wr  
            dWsdt = ( (alphaS * (cs**sigmaS) )/( (betaS**sigmaS) + (cs**sigmaS) ) ) * Ws

    
        return[dWrdt,dWsdt,dQrdt,dQsdt]

    # integrate the system using RK4 method
    solver = ode(f).set_integrator("dopri5") # vode "dopri5"
    solver.set_initial_value(Y0,t0) 
    
    # Data stores
    tStore = [t0]
    WrStore = [ Y0[0] ]
    WsStore = [ Y0[1] ]
    QrStore = [ Y0[2] ]
    QsStore = [ Y0[3] ]
    crStore = [ cr0 ]
    csStore = [ cs0 ]
            
    # Solve the system, making sure that it is still physically possible
    while solver.successful and solver.t < tfinal \
    and solver.y[0]>=0 and solver.y[1]>=0 and solver.y[2] >=0 and solver.y[3] >=0:
        solver.integrate(solver.t+dt)
        tStore.append(solver.t)
        WrStore.append(solver.y[0])
        WsStore.append(solver.y[1])
        QrStore.append(solver.y[2])
        QsStore.append(solver.y[3])
        crStore.append(solver.y[2]/solver.y[0])
        csStore.append(solver.y[3]/solver.y[1])
    if report==True:
        rgrRoot = math.log( WrStore[len(WrStore)-1] / WrStore[0] ) / tfinal
        rgrShoot = math.log( WsStore[len(WsStore)-1] / WsStore[0] ) / tfinal
        print("Summary: RGR_root = ", round(rgrRoot,3), 
        " RGR_shoot = ", round(rgrShoot,3),
        " Final root mass (g) = ", round(WrStore[len(WrStore)-1],5), 
        " Final shoot mass (g) = ", round(WsStore[len(WsStore)-1],5),
        " Final root conc. = ", round(crStore[len(crStore)-1],2), 
        " Final shoot conc. = ", round(csStore[len(csStore)-1],2))
        if tfinal==63:
            Wflower = WsStore[len(WsStore)-1]
            E = wtG[12]
            N = wtG[13]
            L = wtG[14]
            Y= (E*(Wflower**L))/((Wflower**L)+(N**L))
            print("Yield (g) = ", round(Y,2))
        else:
            None
    else:
        None
    return(tStore, WrStore, WsStore, QrStore, QsStore, crStore, csStore)
    
# tStore, WrStore, WsStore, QrStore, QsStore, crStore, csStore = solver()
    
############################### Quick Plot ####################################
def qplot():
    """quick plot solver outcome"""
    plt.subplot(2,1,1)
    plt.plot(tStore,csStore,"g-", label="Shoot", linewidth=3.0)
    plt.plot(tStore,crStore,"k-", label="Root", linewidth=3.0) 
    plt.xlabel("Time (days)", fontsize=fontSize)
    plt.ylabel(r"Tissue Concentration ($\mu$mol / g FW)", fontsize=fontSize) 
    plt.tick_params(axis='both', which='major', direction="out", labelsize=12,\
                    top="off", right="off")
    plt.ylim(0,210)
    plt.legend(loc="best")
    plt.show()

    plt.plot(tStore,WrStore,"k-", label="Root", linewidth=3.0) 
    plt.plot(tStore,WsStore,"g-", label="Shoot", linewidth=3.0) 
    #plt.plot(tStore, (np.array(WrStore)+np.array(WsStore)*1000),"bo-", label="Plant")
    plt.xlabel("Time (days)", fontsize=fontSize)
    plt.ylabel("Tissue Fresh Weight (g)", fontsize=fontSize)
    plt.tick_params(axis='both', which='major', direction="out", labelsize=12,\
                    top="off", right="off")
    plt.legend(loc=0)
    plt.show()

    
########################### Sensitivity Analysis ##############################
def sa(metric="KUtR", c=limit, divisions=26, tfinal=28, conc=50, concHi=1000, plot=False):   
    """Metrics: rgrRatio, wpRatio, wrRatio, wsRatio, rsr, kupe, qp, qr, qs, qDist, relQp, relQr, relQs,
    cr, cs, relCr, relCs,
    kute, kuse, np, ap, pe, ue, wp, ws, wsL, wr, y, hi
    geo4, geo2"""
    # Data stores
    change =[]
    geneNumber = []
    geneValue = []
    metricScore = []
    for g in range(len(wtG)): # for each gene
        changeMag = []
        geneScore = []
        geneNo = []
        geneVal = []
        for n in decimalRange(-c,(c+(c/divisions)),(2*c)/divisions): # for each of 27 possible values, given limits
            G = mutate(initialGenome=wtG, position=g, by=1+n)
            # low treatment 
            tStore, WrStore, WsStore, QrStore, QsStore, crStore,\
            csStore = solver(G, S=conc, report=False, tfinal=tfinal) # solve system for low external conc.
            QpStore = np.array(QrStore)+np.array(QsStore) # low treatment content store
            WpStore = np.array(WrStore)+np.array(WsStore) # low treatment mass store
            # high treatment
            tStoreHi, WrStoreHi, WsStoreHi, QrStoreHi,QsStoreHi,\
            crStoreHi, csStoreHI = solver(G, S=concHi, report=False, tfinal=tfinal)  # solve system for high external conc.
            QpStoreHi = np.array(QrStoreHi)+np.array(QsStoreHi) # high treatment content store
            WpStoreHi = np.array(WrStoreHi)+np.array(WsStoreHi) # high treatment content store
            if len(tStore) == tfinal + 1: # if a run has been completed for a gene
                changeMag.append(n)                
                geneNo.append(g)
                geneVal.append(G[g])
                if metric == "rgrR": # RGR Ratio
                    rgrLow = math.log( (WpStore[len(WpStore)-1]) / WpStore[0] )   / tfinal
                    rgrHigh = math.log( (WpStoreHi[len(WpStoreHi)-1]) / WpStoreHi[0] )   / tfinal 
                    metricValue = rgrLow/rgrHigh
                elif metric == "RP": # whole plant mass ratio
                    metricValue = WpStore[len(WpStore)-1] / WpStoreHi[len(WpStoreHi)-1]
                elif metric == "RR": # root mass ratio
                    metricValue = WrStore[len(WrStore)-1] / WrStoreHi[len(WrStoreHi)-1]
                elif metric == "RS": # shoot mass ratio
                    metricValue = WsStore[len(WsStore)-1] / WsStoreHi[len(WsStoreHi)-1]
                elif metric == "rsr":
                    metricValue = WrStore[len(WrStore)-1] / WsStore[len(WsStore)-1]
                elif metric == "U": # uptake efficiency
                    metricValue = QpStore[len(QpStore)-1] / WrStore[len(WrStore)-1]
                elif metric == "QP": # plant content
                    metricValue = QpStore[len(QpStore)-1]
                elif metric == "QR": # rot K contnet
                    metricValue = QrStore[len(QrStore)-1]
                elif metric == "QS": # shoot K content
                    metricValue = QsStore[len(QsStore)-1]
                elif metric == "qDist":
                    metricValue = QrStore[len(QrStore)-1]/QsStore[len(QsStore)-1]
                elif metric == "RQP": # relative plant K content
                    metricValue = QpStore[len(QpStore)-1] / QpStoreHi[len(QpStoreHi)-1]
                elif metric == "RQR": # relative root K content
                    metricValue = QrStore[len(QrStore)-1] / QrStoreHi[len(QrStoreHi)-1]
                elif metric == "RQS": # relative shoot K content
                    metricValue = QsStore[len(QsStore)-1] / QsStoreHi[len(QsStoreHi)-1]
                elif metric == "CR": # root K conc
                    metricValue = QrStore[len(QrStore)-1] / WrStore[len(WrStore)-1]
                elif metric == "CS": # shoot K conc
                    metricValue = QsStore[len(QsStore)-1] / WsStore[len(WsStore)-1]
                elif metric == "RCR": # relative root K conc
                    metricValue = (QrStore[len(QrStore)-1] / WrStore[len(WrStore)-1]) / (QrStoreHi[len(QrStoreHi)-1] / WrStoreHi[len(WrStoreHi)-1])
                elif metric == "RCS": # realtive shoot K conc
                    metricValue = (QsStore[len(QsStore)-1] / WsStore[len(WsStore)-1]) / (QsStoreHi[len(QsStoreHi)-1] / WsStoreHi[len(WsStoreHi)-1])
                elif metric == "KUtR": # K utilisation ratio
                    metricValue = (WsStore[len(WsStore)-1] / QsStore[len(QsStore)-1])
                elif metric == "KUtI": # K utilisation index
                    metricValue = (WsStore[len(WsStore)-1] / csStore[len(csStore)-1])
                elif metric == "NP": #nutrient productivity
                    metricValue = (WsStore[len(WsStore)-1] - WsStore[0]) / QsStore[len(QsStore)-1]
                elif metric == "AP": # accumulative productivity
                    metricValue = (math.log(WsStore[len(WsStore)-1] / WsStore[0]))/ ( (QsStore[len(QsStore)-1]/WsStore[len(WsStore)-1]) * tfinal )
                elif metric == "PE": # physiological efficiency
                    wsHigh = WsStoreHi[len(WsStoreHi)-1]
                    ws = WsStore[len(WsStore)-1]
                    qsHigh = QsStoreHi[len(QsStoreHi)-1]
                    qs = QsStore[len(QsStore)-1]
                    metricValue = (wsHigh-ws)/(qsHigh-qs)
                #elif metric == "ue":
                    #metricValue = WpStore[len(WpStore)-1] / ((QpStore[len(QpStore)-1]/WpStore[len(WpStore)-1]))
                elif metric == "P": # plant mass
                    metricValue = WpStore[len(WpStore)-1]
                elif metric == "R": # root mass
                    metricValue = WrStore[len(WrStore)-1]
                elif metric == "S": # shoot mass
                    metricValue = WsStore[len(WsStore)-1]
                elif metric == "PL": # plant mass long term
                    tStore, WrStore, WsStore, QrStore, QsStore, crStore,\
                    csStore = solver(G, S=conc, report=False, tfinal=65) # solve system for low external conc.
                    QpStore = np.array(QrStore)+np.array(QsStore) # low treatment content store
                    WpStore = np.array(WrStore)+np.array(WsStore) # low treatment mass store
                    metricValue = WpStore[len(WpStore)-1]
                elif metric == "SL": # shoot mass long term
                    tStore, WrStore, WsStore, QrStore, QsStore, crStore,\
                    csStore = solver(G, S=conc, report=False, tfinal=65) # solve system for low external conc.
                    QpStore = np.array(QrStore)+np.array(QsStore) # low treatment content store
                    WpStore = np.array(WrStore)+np.array(WsStore) # low treatment mass store
                    metricValue = WsStore[len(WsStore)-1]
                elif metric == "RL": # plant mass long term
                    tStore, WrStore, WsStore, QrStore, QsStore, crStore,\
                    csStore = solver(G, S=conc, report=False, tfinal=65) # solve system for low external conc.
                    QpStore = np.array(QrStore)+np.array(QsStore) # low treatment content store
                    WpStore = np.array(WrStore)+np.array(WsStore) # low treatment mass store
                    metricValue = WrStore[len(WrStore)-1]
                elif metric == "GY": # grain yield
                    # low treatment - yield
                    tStore, WrStore, WsStore, QrStore, QsStore, crStore,\
                    csStore = solver(G, S=conc, report=False, tfinal=65) # solve system for low external conc.
                    QpStore = np.array(QrStore)+np.array(QsStore) # low treatment content store
                    WpStore = np.array(WrStore)+np.array(WsStore) # low treatment mass store
                    Wflower = WsStore[len(WsStore)-1]
                    E = G[12]
                    N = G[13]
                    L = G[14]
                    metricValue = (E*(Wflower**L))/((Wflower**L)+(N**L))
                elif metric == "HI": # harvest index
                    tStore, WrStore, WsStore, QrStore, QsStore, crStore,\
                    csStore = solver(G, S=conc, report=False, tfinal=65) # solve system for low external conc.
                    QpStore = np.array(QrStore)+np.array(QsStore) # low treatment content store
                    WpStore = np.array(WrStore)+np.array(WsStore) # low treatment mass store
                    Wflower = WsStore[len(WsStore)-1]
                    E = G[12]
                    N = G[13]
                    L = G[14]
                    metricValue = (E*(Wflower**L))/((Wflower**L)+(N**L)) / ((E*(Wflower**L))/((Wflower**L)+(N**L)) + Wflower)
                    #print(Vflower, metricValue)
                elif metric == "G4":
                    ws = WsStore[len(WsStore)-1]   
                    wr = WrStore[len(WrStore)-1]
                    cr = QrStore[len(QrStore)-1] / WrStore[len(WrStore)-1]
                    cs = QsStore[len(QsStore)-1] / WsStore[len(WsStore)-1]
                    metricValue = (ws*wr*cr*cs)**(1/4)
                elif metric == "G2":
                    ws = WsStore[len(WsStore)-1] 
                    cs = QsStore[len(QsStore)-1] / WsStore[len(WsStore)-1]
                    metricValue = (ws*cs)**(1/2)
                elif metric == "GUKUtI":
                    U = QpStore[len(QpStore)-1] / WrStore[len(WrStore)-1]
                    KUtI = (WpStore[len(WpStore)-1] / csStore[len(csStore)-1])
                    metricValue = (U*KUtI)**(1/2)
                else: 
                    return("Error: no metric assigned")
                geneScore.append(metricValue) # add calcualted metric value to data store
            else:
                changeMag.append(n)
                geneNo.append(g)
                geneVal.append(G[g])
                geneScore.append("NaN") # reset the system for the next gene
        change.append(changeMag)        
        metricScore.append(geneScore)
        geneNumber.append(geneNo)
        geneValue.append(geneVal)
    
    if plot==True:
        for g in range(len(wtG)): # for each gene
            plt.plot([x*100 for x in change[0]], metricScore[g]) # first bit scales x axis
        plt.show()
    else:
        None
    
    return(change, geneNumber, geneValue, metricScore)

# run with: c, gn, gv, ms = sa(metric="GY")

####################### Elasticity Analysis ###################################
def ea(geneValue=None, metricScore=None):
    """Run sensitivity analysis with c, gn, gv, ms = sa() first"""
    geneNo = []
    elasticity = []
    for g in range(len(wtG)):
        p = geneValue[g][13]
        pStar = geneValue[g][26]
        m = metricScore[g][13]
        mStar = metricScore[g][26]
        elast = ((mStar-m)/(pStar-p))*(p/m)
        geneNo.append(g)
        elasticity.append(elast)
    plt.bar(geneNo, elasticity, color="black")
    locs, labels = plt.xticks(list(range(0,10)), [params[i] for i in range(len(params))])
    plt.tick_params(axis="x", labelsize=16)
    plt.xticks([0.5 + i for i in range(len(params))])
    plt.xlabel("Constant", fontsize=fontSize)
    plt.ylabel("Standardised Sensitivity", fontsize=fontSize)
    plt.tick_params(axis='both', which='major', direction="out", \
                    top="off", right="off")
    plt.plot()
    plt.show()
    return(geneNo, elasticity)

# run with: gno, e = ea(geneValue=gv, metricScore=ms)


######################## Automated analysis ################################### 
f = ["GY","HI","P","S","R","RP","RS","RR","PL","SL","RL",
        "CS","CR","RCS","RCR","KUtR","KUtI","NP","AP","PE",
        "QP","QS","QR","RQP","RQS","RQR","U"]#,"G2","G4","GUKUtI"]

def an(metric="full", tfinal=28, conc=100, concHi=1000):
    """Analyse a specified trait"""
    if metric=="full":
        for i in range(len(f)):
            print(f[i])
            c, gn, gv, ms = sa(metric=f[i], tfinal=tfinal, conc=conc, concHi=concHi)
            gno, e = ea(geneValue=gv, metricScore=ms)
            fileName = str(f[i])+str("_")+str(tfinal)+str("_Days_")+str(conc)+str("_uM")+str(".csv")
            wCSV(geneNo=gno, elast=e, name=fileName)    
    
    else:
        c, gn, gv, ms = sa(metric=metric, tfinal=tfinal, conc=conc, concHi=concHi)
        gno, e = ea(geneValue=gv, metricScore=ms)
        fileName = str(metric)+str("_")+str(tfinal)+str("_Days_")+str(conc)+str("_uM")+str(".csv")
        wCSV(geneNo=gno, elast=e, name=fileName)
# run with: an(tfinal=28, conc=50) 
    
